#!/usr/bin/tclsh
## July '97
## This script is a metaserver for tkbc users. It maintains a list
## of active users which is modified and retrieved via a tcp/ip
## connection. The user side of tkbc periodically updates it's entry
## in this server's database so it isn't aged out of the list.

## the network interface as encountered by the client or user side of tkbc
## will be as follows:
## get
##    returns servername, version, user_list
## add name port cookie
## del name cookie
## update name cookie

## the configuration file for the meataserver contains the following
## information.
## name xxxxxxxxxx
## port xxxx
## ageing_interval seconds
## entry_life aging_intervals
##
## which might look like this:
## name "TkBC Zone 1"
## port 5005
## age_int 30
## entry_life 3

## global variables
set version "3.0b1"
set rcfilepath $env(HOME)/.tkbcmsrc
## file handle
set rcfile ""
set myport 5000
set myname "TkBC Metaserver"
set age_int 60
set entry_life 2

set user_list(dummy) dum
unset user_list(dummy)

##--------------------------------------
## config file stuff
##--------------------------------------

proc findRc {} {
    ## check for the .tkbcmsrc file
    global rcfilepath rcfile
    if { [file exists $rcfilepath] } {
	set rcfile [open $rcfilepath r]
    } else {
	puts "$rcfilepath missing. please configure before running"
	exit
    }
}

##############################
## Age registered entries
##############################

proc ageEntries {}  {
    global user_list age_int entry_life

    foreach entry [array names user_list] {
	set now [clock seconds]
	set age [expr $now - [lindex $user_list($entry) 3]]
	if { $age > [expr $age_int * $entry_life] } {
	    unset user_list($entry)
	}
    }

    after $age_int {
	ageEntries
    }   
}

####################################
## Callback to handle connections
####################################

proc gotClient { socket machine port } {
    ## we got a connection. print out version information and wait for
    ## a command. if we don't get anything in a reasonable about of time
    ## shut down the socket.

    global version myport myname entry_life age_int user_list
    #puts "got connection on $socket from $machine $port"
    
    ## we are not going to allow the client to specify the machine
    ## for security reasons.

    puts $socket "TkBC Metaserver $version on port $myport with age interval \
	    of $age_int seconds and entry life of $entry_life \
	    called $myname ready."
    flush $socket
  
    ## get the command
    gets $socket line

    ## make sure there are no nasty escape chars or commands in the line
    if  { [regexp {(\{|\}|exec|\[|\]|\$)} $line] } {
	puts $socket "ERROR-04 illegal characters in command"
	flush $socket
	close $socket
	puts "illegal characters in command from $machine"
    } else {
	
	## get returns user_list
	## add name port cookie
	## del name cookie
	## update name cookie
	
	## GET command
	if { [regexp {^get.*} $line] } {
	    puts $socket "OK"
	    foreach entry [array names user_list] {
		puts $socket "\"$entry\" [lrange $user_list($entry) 0 1]"
		flush $socket
	    }
	    flush $socket
	    close $socket
	} elseif { [regexp {^add\ .*} $line] } {
	    ## ADD command
	    ## need to add test for proper number of elements
	    set newname [lindex $line 1]
	    set newport [lindex $line 2]
	    set newcookie [lindex $line 3]
	    set dupflag 0
	    foreach entry [array names user_list] {
		## if it already exits return error
		if { [string compare $entry $newname] == 0 } {
		    set dupflag 1
		    puts $socket "ERROR-01 entry $newname already exists"
		    flush $socket
		    close $socket
		}
	    }
	    ## if it doesn't exist, add it
	    if { !$dupflag } {
		set user_list($newname) \
			"$machine $newport $newcookie [clock seconds]"
		puts $socket "OK"
		flush $socket
		close $socket
	    }
	} elseif { [regexp {^update\ .*} $line] } {
	    ## UPDATE command
	    set thename [lindex $line 1]
	    
	    set thecookie [lindex $line 2]
	    ## if the cookies match, do an update
	    if { [string compare $thecookie [lindex $user_list($thename) 2]] == 0 } {
		set user_list($thename) \
			"[lrange $user_list($thename) 0 2] [clock seconds]"
		puts $socket "OK"
		flush $socket
		close $socket
	    } else {
		puts $socket "ERROR-02 mismatched cookies for $thename"
		flush $socket
		close $socket
	    }
	} elseif { [regexp {^del\ .*} $line] } {
	    ## DEL command
	    set thename [lindex $line 1]
	    set thecookie [lindex $line 2]
	    ## if the cookies match, do an update
	    if { [string compare $thecookie [lindex $user_list($thename) 2]] == 0 } {
		unset user_list($thename)
		puts $socket "OK"
		flush $socket
		close $socket
	    } else {
		puts $socket "ERROR-02 mismatched cookies for $thename"
		flush $socket
		close $socket
	    }
	}
    }
}

##---------------------------------------
## begin execution
##---------------------------------------

## read in the rc file and set things up
## the rcfile will be of the format described above

if {$argc > 0} {
    set rcfilepath [lindex $argv 0]
}

findRc

while { [gets $rcfile line] != -1 } {
    if { [regexp {^name\ .*} $line] } {
	set myname [lindex $line 1]
    }
    if { [regexp {^port\ .*} $line] } {
	set myport [lindex $line 1]
    }
    if { [regexp {^age_int\ .*} $line] } {
	set age_int [lindex $line 1]
    }
    if { [regexp {^entry_life\ .*} $line] } {
	set entry_life [lindex $line 1]
    }
    
}
close $rcfile

if { [catch {set sockflag [socket -server { gotClient } $myport]}] != 0 } {
    puts "port busy or other error. Exiting...."
    exit
}

## schedule entry aging
after $age_int {
    ageEntries
}

vwait yermom
